/**********************************************************************\
 © COPYRIGHT 2015 Corporation for National Research Initiatives (CNRI);
                        All rights reserved.

        The HANDLE.NET software is made available subject to the
      Handle.Net Public License Agreement, which may be obtained at
          http://hdl.handle.net/20.1000/103 or hdl:20.1000/103
\**********************************************************************/

package net.handle.hdllib.trust;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.handle.hdllib.Common;
import net.handle.hdllib.HandleRecord;
import net.handle.hdllib.HandleResolver;
import net.handle.hdllib.HandleValue;
import net.handle.hdllib.Util;

public class HandleRecordTrustVerifier {

    private ChainBuilder chainBuilder;
    private ChainVerifier chainVerifier;
    
    public HandleRecordTrustVerifier(HandleResolver resolver) {
        this.chainBuilder = new ChainBuilder(resolver);
        this.chainVerifier = new ChainVerifier(resolver.getConfiguration().getRootKeys());
    }
    
    public HandleRecordTrustVerifier(ChainBuilder chainBuilder, ChainVerifier chainVerifier) {
        this.chainBuilder = chainBuilder;
        this.chainVerifier = chainVerifier;
    } 
    
    public boolean validateHandleRecord(HandleRecord handleRecord) {
        List<HandleValue> valuesList = handleRecord.getValues();
        Set<Integer> verifiedValues = new HashSet<Integer>();
        try {
            List<JsonWebSignature> signatures = getJsonWebSignaturesFromValues(handleRecord.getValuesAsArray());
            if (signatures.isEmpty()) return false;
            for (JsonWebSignature jws : signatures) {
                List<IssuedSignature> chain = chainBuilder.buildChain(jws);
                ChainVerificationReport report = chainVerifier.verifyValues(handleRecord.getHandle(), valuesList, chain);
                if (!report.canTrustAndGranted()) {
                    return false;
                }
                if (!report.valuesReport.badDigestValues.isEmpty()) {
                    return false;
                }
                if (!report.valuesReport.missingValues.isEmpty()) {
                    return false;
                }
                verifiedValues.addAll(report.valuesReport.verifiedValues);
            }
        } catch (TrustException e) {
            // TODO should we report this in more detail?
            return false;
        }
        if (isAllValuesThatNeedSignatureInVerifiedSet(handleRecord, verifiedValues)) {
            return true;
        } else {
            return false;
        }
    }

    private List<JsonWebSignature> getJsonWebSignaturesFromValues(HandleValue[] newValues) throws TrustException {
        HandleValue[] hsSignatureValues = getSignatureValues(newValues);
        List<JsonWebSignature> signatures = new ArrayList<JsonWebSignature>();
        for (HandleValue hsSignatureValue : hsSignatureValues) {
            JsonWebSignature signature = JsonWebSignatureFactory.getInstance().deserialize(hsSignatureValue.getDataAsString());
            signatures.add(signature);
        }
        return signatures;
    }

    private static boolean isAllValuesThatNeedSignatureInVerifiedSet(HandleRecord newRecord, Set<Integer> verifiedValues) {
        for (HandleValue value : newRecord.getValues()) {
            int index = value.getIndex();
            if (!verifiedValues.contains(index)) {
                if (valueNeedsSignature(newRecord.getHandle(), value)) {
                    return false;
                }
            }
        }
        return true;
    }
    
    @SuppressWarnings("deprecation")
    private static boolean valueNeedsSignature(String handle, HandleValue value) {
        if (value.hasType(Common.HS_SIGNATURE_TYPE)) return false;
        if (!value.getAnyoneCanRead()) return false;
        if ("0.NA/0.NA".equalsIgnoreCase(handle) && (value.hasType(net.handle.hdllib.HandleSignature.SIGNATURE_TYPE) || value.hasType(net.handle.hdllib.HandleSignature.METADATA_TYPE))) return false;
        return true;
    }

    private static HandleValue[] getSignatureValues(HandleValue[] newValues) {
        return Util.filterValues(newValues, null, Common.HS_SIGNATURE_TYPE_LIST);
    }
}
